// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_AncRx.cpp *#*#*#*#*#*#*#*#*#* (C) 2025 DekTec
//
// Example: AvFifo_AncRx  SMPTE 2110-40 ANC Receiver
//
// This example shows how to receive SMPTE 2110-40 ANC packets using the AvFifo API.
// It connects to a fixed IP and port, receives the ANC data, and prints key metadata.
// 
// Configuration is done through a set of constants below.
// To adapt the example to your setup, update the constants and recompile.
//
// Note:
//   This example focuses on receiving SMPTE 2110-40 ANC data items.
//   For a more complete overview of AvFifo usageincluding device attachment,
//   FIFO lifecycle (start/stop/detach), error handling, and configuration patterns
//   refer to the AvFifo_VideoRx example.
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include required headers -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <chrono>                   // Standard library: timing utilities.
#include <iostream>                 // Standard library: I/O streams.
#include <thread>                   // Standard library: threading support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Device selection.
constexpr int CfgDeviceType{2110};  // Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceNo{0};       // Device number: 0 = first matching device in system.

// IP configuration for the incoming 2110-40 ANC stream.
const AvFifo::IpPars AvFifo_AncRx_IpPars
{
    .IpAddr{239, 40, 0, 1},
    .IpVersion{AvFifo::IpProtocolVersion::IPv4},
    .Port{50040},
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}  // Must be RTP for 2110-40.
};

// ANC stream reception configuration.
// Both decoded user words and raw 10-bit ANC data are captured.
const AvFifo::St2110::RxConfigAnc AvFifo_AncRx_Config
{
    .IncludeUserWords{true},        // Enable reception of decoded user words.
    .IncludeRawData{true}           // Enable reception of raw ANC data.
};

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Forward declarations for top-down function presentation.
void AvFifo_AncRx();
void AncRxLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function prints application details and invokes the AvFifo_AncRx() function.
// Any exceptions are caught and reported.
//
int main()
{
    cout << "\n"
         << "---------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_AncRx - SMPTE 2110-40 ANC receiver\n"
         << "Version " << Version << "\n"
         << "---------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << " device number " << CfgDeviceNo << "\n"
         << "Expected IP address: "
         << (int)AvFifo_AncRx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_AncRx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_AncRx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_AncRx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_AncRx_IpPars.Port << "\n"
         << "Press Ctrl+C to stop the application.\n\n";

    try {
        AvFifo_AncRx();
    }
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e)          {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                   {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_AncRx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Opens the DekTec device, checks network readiness, runs the main ANC receive loop,
// and finally detaches from the hardware.
//
void AvFifo_AncRx()
{
    DtDevice Device{};
    DTAPI_RESULT Res = Device.AttachToType(CfgDeviceType, CfgDeviceNo);
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    // Verify that network port 1 is operational.
    Res = Device.IsNetworkCardOperational(1, 
                          AvFifo_AncRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                          AvFifo_AncRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    try {
        AncRxLoop(Device);
    }
    catch (AvFifo::DriverError& e) {cerr << "Driver error: " << e.what() << '\n';}

    Device.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AncRxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Core function of the example.
// Attaches an RxFifo to the device, configures reception, and displays ANC packet info.
//
void AncRxLoop(DtDevice& Device)
{
    AvFifo::St2110::RxFifoAnc AncRxFifo{};
    AncRxFifo.Attach(Device, 1, AvFifo::HwOrSwPipe::UseSwPipe);

    AncRxFifo.SetIpPars(AvFifo_AncRx_IpPars);
    AncRxFifo.Configure(AvFifo_AncRx_Config);
    
    if (!AncRxFifo.Start())
    {
        AvFifo::FifoStatus Status{AncRxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        AncRxFifo.Detach();
        return;
    }

    while (true)
    {
        // Check FIFO status.
        AvFifo::FifoStatus Status{AncRxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // Process next available ANC frame, if any.
        if (AncRxFifo.GetFifoLoad() > 0)
        {
            AvFifo::St2110::RxAncFrame* Frame{AncRxFifo.Read()};

            cout << "--------------------------------------------\n";
            cout << "Received ANC data: #Packets: " << Frame->Packets.size() << '\n';
            cout << "--------------------------------------------\n";

            for (AvFifo::St2110::RxAncPacket Packet: Frame->Packets)
            {
                cout << "DID: " << static_cast<int>(Packet.Id.Did) << '\n';
                if (Packet.Id.Sdid)
                    cout << "SDID: " << static_cast<int>(Packet.Id.Sdid.value()) << '\n';
                if (Packet.Dbn)
                    cout << "DBN: " << static_cast<int>(Packet.Dbn.value()) << '\n';
                if (Packet.StreamNumber)
                    cout << "StreamNumber: " << Packet.StreamNumber.value() << '\n';
                cout << "LineNumber: " << Packet.Location.LineNumber << '\n';
                cout << "HorizontalOffset: " << Packet.Location.HorizontalOffset << '\n';
                if (Packet.Location.IsColorDiff)
                    cout << "ANC packet located in the color difference data channel\n";
                if (Packet.UserWords)
                    cout << "#UserWord elements: " << Packet.UserWords.value().size() << '\n';
                if (Packet.RawData)
                    cout << "#RawData elements: " << Packet.RawData.value().size() << '\n';
                cout << "--------------------------------------------\n";
            }
            AncRxFifo.ReturnToMemPool(Frame);
        } else {
            // No frame available; sleep briefly and try again.
            this_thread::sleep_for(10ms);
        }
    }

    // Stop and clean up after loop ends.
    AncRxFifo.Stop();
    AncRxFifo.Detach();
}
