// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_AncTx.cpp *#*#*#*#*#*#*#*#*#* (C) 2025 DekTec
//
// Example: AvFifo_AncTx  SMPTE 2110-40 ANC Transmitter
// 
// This example shows how to transmit SMPTE 2110-40 ANC packets using the AvFifo API.
// It periodically sends a test ANC packet over the network, alternating between
// raw ANC and SDI-style packets.
//
// The ANC packet content is synthetic and for test purposes only. It does not
// represent any actual ANC metadata format.
// 
// Configuration is done through a set of constants below.
// To adapt the example to your setup, update the constants and recompile.
// 
// Note:
//   This example focuses on transmitting ANC data frames with synthetic test data.
//   For a more complete overview of AvFifo usageincluding device attachment,
//   FIFO lifecycle (start/stop/detach), error handling, and configuration patterns
//   refer to the AvFifo_VideoTx example.
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include required headers -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <chrono>                   // Standard library: timing utilities.
#include <iostream>                 // Standard library: I/O streams.
#include <thread>                   // Standard library: threading support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Device selection.
constexpr int CfgDeviceType{2110};  // Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceNo{0};       // Device number: 0 = first matching device in system.

// IP configuration for outgoing 2110-40 ANC stream.
const AvFifo::IpPars AvFifo_AncTx_IpPars
{
    .IpAddr{239, 40, 0, 1},
    .IpVersion{AvFifo::IpProtocolVersion::IPv4},
    .Port{50040},
    .RtpPayloadType{100},
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}
};

// ANC stream transmission configuration.
const AvFifo::St2110::TxConfigAnc AvFifo_AncTx_Config
{
    .VideoScanning{AvFifo::St2110::VideoScanning::Progressive}
};

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Forward declarations for top-down function presentation.
void AvFifo_AncTx();
void AncTxLoop(DtDevice&);
vector<uint16_t> GenerateRawAncData(uint8_t Did, uint8_t Sdid,
                                                   const vector<uint16_t>& UserDataWords);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function prints application details and invokes the AvFifo_AncTx() function.
// Any exceptions are caught and reported.
//
int main()
{
    cout << "\n"
         << "------------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_AncTx - SMPTE 2110-40 ANC transmitter\n"
         << "Version " << Version << "\n"
         << "------------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << " device number " << CfgDeviceNo << "\n"
         << "Destination IP address: "
         << (int)AvFifo_AncTx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_AncTx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_AncTx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_AncTx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_AncTx_IpPars.Port << "\n"
         << "Press Ctrl+C to stop the application.\n\n";

    try {
        AvFifo_AncTx();
    }
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e)          {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                   {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_AncTx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Opens the DekTec device, checks network readiness, runs the main ANC transmit loop,
// and finally detaches from the hardware.
//
void AvFifo_AncTx()
{
    DtDevice Device{};
    DTAPI_RESULT Res = Device.AttachToType(CfgDeviceType, CfgDeviceNo);
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    // Verify that network port 1 is operational.
    Res = Device.IsNetworkCardOperational(1, 
                          AvFifo_AncTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                          AvFifo_AncTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    try {
        AncTxLoop(Device);
    }
    catch (AvFifo::DriverError& e)     {cerr << "Driver error: " << e.what() << '\n';}
    catch (AvFifo::SchedulingError& e) {cerr << e.what() << '\n';
    }

    Device.Detach();
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AncTxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Core function of the example.
// Initializes and configures the TxFifo for SMPTE 2110-40 ANC transmission, and
// continuously transmits alternating ANC packet formats.
// 
void AncTxLoop(DtDevice& Device)
{
    AvFifo::St2110::TxFifoAnc AncTxFifo{};
    AncTxFifo.Attach(Device, 1, AvFifo::HwOrSwPipe::UseSwPipe);

    AncTxFifo.SetIpPars(AvFifo_AncTx_IpPars);
    AncTxFifo.Configure(AvFifo_AncTx_Config);

    // Set the initial Time-of-Day (ToD) slightly in the future (now + 100ms) to ensure
    // there is time to prepare the first frames before their intended Tx timestamp.
    DtTimeOfDay ToD{};
    Device.GetTimeOfDay(ToD);
    ToD += 100'000'000;             // 100,000,000ns = 100ms

    // Start transmission; return false means no data can be sent (e.g. IP misconfigured).
    if (!AncTxFifo.Start())
    {
        AvFifo::FifoStatus Status{AncTxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        AncTxFifo.Detach();
        return;
    }

    // Main loop: keep transmitting until a FIFO failure occurs.
    bool UseSdiAnc = false;
    while (true)
    {
        // Check if the FIFO is still healthy and transmitting.
        AvFifo::FifoStatus Status{AncTxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // Proceed only if there is space in the FIFO.
        if (AncTxFifo.GetFifoLoad() < AncTxFifo.GetMaxSize())
        {
            // Align ToD to the 50-Hz media grid.
            ToD = AvFifo::Tod2Grid_Video(ToD, {50, 1});

            // Allocate a new frame from the memory pool.
            AvFifo::St2110::TxAncFrame* Frame = AncTxFifo.GetFrameFromMemPool();
            Frame->ToD = ToD;
            Frame->RtpTime = AvFifo::St2110::Tod2Rtp_Video(ToD);
            Frame->Field = 0;

            // Define the ANC packet metadata (position and stream).
            AvFifo::St2110::TxAncPacket Packet;
            Packet.Location.HorizontalOffset = 1928;
            Packet.Location.IsColorDiff = false;
            Packet.Location.LineNumber = 10;
            Packet.StreamNumber = 1;

            // Define ANC payload content.
            const uint8_t Did = 0x60;
            const uint8_t SDid = 0x60;
            vector<uint16_t> UserWords{
                                  0x110, 0x228, 0x200, 0x200, 0x140, 0x200, 0x200, 0x200,
                                  0x200, 0x200, 0x200, 0x200, 0x110, 0x200, 0x200, 0x200};

            // Alternate between two ways of constructing the ANC payload.
            if (UseSdiAnc)
            {
                // SDI-style packet using DID/SDID and user words.
                AvFifo::St2110::TxAncPacket_Sdi SdiAnc{.Did=Did, .SdidOrDbn=SDid};
                SdiAnc.UserWords = UserWords;
                Packet.Packet = SdiAnc;
            } else {
                // Raw 10-bit packet with explicit encoding and checksum.
                Packet.Packet = GenerateRawAncData(Did, SDid, UserWords);
            }

            // Add the packet to the frame and queue it for transmission.
            Frame->Packets.push_back(Packet);
            AncTxFifo.Write(Frame);

            // Toggle between SDI and RAW formats.
            UseSdiAnc = !UseSdiAnc;
		    
		    // Advance ToD to the next frame time (20ms later for 50-Hz video).
            ToD += 20'000'000;      // 20,000,000ns = 20ms
        } else {
            // FIFO full; sleep briefly to reduce CPU usage.
            this_thread::sleep_for(10ms);
        }
    }

    // Clean up the FIFO and detach from hardware.
    AncTxFifo.Stop();
    AncTxFifo.Detach();
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Gen10Bit -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Converts an 8-bit value into a 10-bit SDI-style word with parity encoding.
//
// The 10-bit word consists of:
//   - Bit 8: even parity over the 8 LSBs (bits 7-0)
//   - Bit 9: inverse of the parity bit (bit 8)
//   - Bits 7-0: original 8-bit value
//
uint16_t Gen10Bit(uint8_t Value)
{
    // Compute even parity over bits 7-0.
    uint16_t Parity = 0;
    for (int i = 0; i < 8; ++i)
        Parity ^= (Value >> i) & 1;

    // Bit 9 = inverted parity bit.
    uint16_t InvParity = Parity ^ 1;

    // Assemble 10-bit value: [InvParity | Parity | 8-bit value]
    return (InvParity << 9) | (Parity << 8) | Value;
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- GenerateRawAncData -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Constructs a RAW ANC packet from SDI-style components and user data.
//
// The generated packet consists of 10-bit words:
//   [DID] [SDID] [DataCount] [UserWords...] [Checksum]
//
// Each component is parity-encoded using Gen10Bit(). The checksum is computed over the
// lower 9 bits of all preceding words and then extended to 10 bits.
//
vector<uint16_t> GenerateRawAncData(uint8_t Did, uint8_t Sdid, 
                                                    const vector<uint16_t>& UserDataWords)
{
    // Encode DID, SDID, and user-word count using SDI parity format.
    uint16_t Did10b = Gen10Bit(Did);
    uint16_t SDid10b = Gen10Bit(Sdid);
    uint16_t DataCount10b = Gen10Bit((uint8_t)UserDataWords.size());

    // Initialize RAW ANC data with preamble: DID, SDID, COUNT.
    vector<uint16_t> AncRawData{Did10b, SDid10b, DataCount10b};

    // Append user words to the raw packet.
    AncRawData.insert(AncRawData.end(), UserDataWords.begin(), UserDataWords.end());

    // Compute checksum over all preceding words (9-bit only, bit 9 ignored).
    uint16_t Sum = (Did10b & 0x1ff) + (SDid10b & 0x1ff) + (DataCount10b & 0x1ff);
    
    // Add sum of all UserData values (ignore bit 9)
    for (uint16_t Value : UserDataWords)
        Sum += Value & 0x1FF;

    // Final checksum value: sum of all words, masked to 9 bits.
    Sum &= 0x1FF;

    // Construct 10-bit checksum: bit 9 = inverse of bit 8.
    uint16_t Checksum = (Sum & 0x1FF) | ((~Sum & 0x100) << 1);
    
    // Append checksum to the packet.
    AncRawData.push_back(Checksum);
    return AncRawData;
}
