// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_AudioRx.cpp *#*#*#*#*#*#*#*#*# (C) 2025 DekTec
//
// Example: AvFifo_AudioRx  SMPTE 2110-30 Audio Receiver
//
// This example shows how to receive SMPTE 2110-30 audio packets using the AvFifo API. It
// connects to a fixed IP and port, receives the audio data, and renders it using SDL.
// 
// Configuration is done through a set of constants below.
// To adapt the example to your setup, update the constants and recompile.
//
// Note:
//   This example focuses on receiving SMPTE 2110-30 PCM audio data.
//   For a more complete overview of AvFifo usageincluding device attachment,
//   FIFO lifecycle (start/stop/detach), error handling, and configuration patterns
//   refer to the AvFifo_VideoRx example.
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// -.-.-.-.-.-.-.-.-.-.-.- Include necessary headers and libraries -.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <SDL.h>                    // Cross-platform library providing low-level access 
                                    // to video, audio and other I/O functions.
#include <chrono>                   // Standard library: timing utilities.
#include <iostream>                 // Standard library: I/O streams.
#include <span>                     // Standard library: std::span.
#include <thread>                   // Standard library: threading support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Device selection.
constexpr int CfgDeviceType{2110};  // Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceNo{0};       // Device number: 0 = first matching device in system. 

// IP configuration for the incoming 2110-30 Audio stream.
const AvFifo::IpPars AvFifo_AudioRx_IpPars
{
    .IpAddr{239, 30, 0, 1},
    .IpVersion{AvFifo::IpProtocolVersion::IPv4},
    .Port{50030},
    .RtpPayloadType{96},
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}  // Must be RTP for 2110-30.
};

// Configuration for receiving 24-bit audio.
const AvFifo::St2110::RxConfigAudio AvFifo_AudioRx_Config
{
    .Format{AvFifo::St2110::AudioFormat::L24BE},
    .SampleRate{48000}      // Defines the audio sample rate in Hz.
};

constexpr int NumChannels{2};

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Stores the SDL state and takes care of proper initialization/destruction.
struct SdlContext
{
    SdlContext();                     // Initializes SDL.
    ~SdlContext();                    // Cleans up SDL.
    void RenderAudio(AvFifo::Frame*); // Renders an audio frame.

    SDL_AudioDeviceID AudioDevice{0};

    bool Paused{true};
    uint32_t MinQueuedAudioSize{0};
};

// Forward declarations for top-down function presentation.
void AvFifo_AudioRx();
void AudioRxLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function prints application details and invokes the AvFifo_AudioRx() function.
// Any exceptions are caught and reported.
//
int main(int argc, char* argv[])
{
    cout << "\n"
         << "-------------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_AudioRx - SMPTE 2110-30 Audio receiver\n"
         << "Version " << Version << "\n"
         << "-------------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << " device number " << CfgDeviceNo << "\n"
         << "Expected IP address: "
         << (int)AvFifo_AudioRx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_AudioRx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_AudioRx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_AudioRx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_AudioRx_IpPars.Port << "\n"
         << "Press Ctrl+C to stop the application.\n\n";

    try {
        AvFifo_AudioRx();
    }
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e)          {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                   {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_AudioRx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Opens the DekTec device, checks network readiness, runs the main receive loop,
// and finally detaches from the hardware.
//
void AvFifo_AudioRx()
{
    DtDevice Device{};
    DTAPI_RESULT Res = Device.AttachToType(CfgDeviceType, CfgDeviceNo);
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    Res = Device.IsNetworkCardOperational(1, 
                        AvFifo_AudioRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                        AvFifo_AudioRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    try {
        AudioRxLoop(Device);
    }
    catch (AvFifo::DriverError& e) {cerr << "Driver error: " << e.what() << '\n';}

    Device.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AudioRxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Core function of the example.
// Attaches an RxFifo to the device, configures reception, and renders the audio using
// SDL.
//
void AudioRxLoop(DtDevice& Device)
{
    AvFifo::RxFifo AudioRxFifo{};
    AudioRxFifo.Attach(Device, 1);

    AudioRxFifo.SetIpPars(AvFifo_AudioRx_IpPars);
    AudioRxFifo.Configure(AvFifo_AudioRx_Config);
    AudioRxFifo.SetMaxSize(400);
    
    // Initialize SDL for displaying the received video frames.
    SdlContext Sdl{};

    if (!AudioRxFifo.Start())
    {
        AvFifo::FifoStatus Status{AudioRxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        AudioRxFifo.Detach();
        return;
    }

    while (true)
    {
        // Check FIFO status.
        AvFifo::FifoStatus Status{AudioRxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // Process next available frame, if any.
        if (AudioRxFifo.GetFifoLoad() > 0)
        {
            AvFifo::Frame* Frame{AudioRxFifo.Read()};
            Sdl.RenderAudio(Frame);
            AudioRxFifo.ReturnToMemPool(Frame);
        } else {
            // No frame available; sleep briefly and try again.
            this_thread::sleep_for(10ms);
        }
    }

    // Stop and clean up after loop ends.
    AudioRxFifo.Stop();
    AudioRxFifo.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::SdlContext -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Initializes SDL for audio rendering.
//
SdlContext::SdlContext()
{
    if (SDL_Init(SDL_INIT_AUDIO) < 0)
        throw runtime_error("SDL initialization failed: " + string(SDL_GetError()));

    SDL_AudioSpec ReqAudioSpec;
    SDL_zero(ReqAudioSpec);
    ReqAudioSpec.freq = AvFifo_AudioRx_Config.SampleRate;
    ReqAudioSpec.format = AUDIO_S16MSB; // 16-bit big endian
    ReqAudioSpec.channels = NumChannels;
    ReqAudioSpec.samples = ReqAudioSpec.channels * 2048;

    SDL_AudioSpec AudioSpec;
    SDL_zero(AudioSpec);
    
    AudioDevice = SDL_OpenAudioDevice(nullptr, 0, &ReqAudioSpec, &AudioSpec, 0);
    if (AudioDevice <= 0)
        throw runtime_error("SDL_OpenAudioDevice failed: " + string(SDL_GetError()));

    Paused = true;
    MinQueuedAudioSize = (ReqAudioSpec.freq * ReqAudioSpec.channels * 2) / 10; // 100ms
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::~SdlContext -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Release the SDL resources. 
//
SdlContext::~SdlContext()
{
    if (AudioDevice)
    {
        SDL_CloseAudioDevice(AudioDevice);
        AudioDevice = 0;
    }
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::RenderAudio -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Renders an AvFifo frame using SDL.
//
void SdlContext::RenderAudio(AvFifo::Frame* Frm)
{
    span<uint8_t> AudioData;

    if (AvFifo_AudioRx_Config.Format == AvFifo::St2110::AudioFormat::L16BE)
        AudioData = span<uint8_t>{Frm->Data(), Frm->Size()};
    else if (AvFifo_AudioRx_Config.Format == AvFifo::St2110::AudioFormat::L24BE)
    {
        static vector<uint8_t> Temp;

        int NSamples{(int)Frm->Size() / 3};
        Temp.resize((size_t)NSamples * 2);

        const uint8_t* Src{Frm->Data()};
        uint8_t* Dst{Temp.data()};

        for (int i{0}; i < NSamples; i++)
        {
            Dst[0] = Src[0];
            Dst[1] = Src[1];
            Src += 3;
            Dst += 2;
        }

        AudioData = Temp;
    }

    if (SDL_QueueAudio(AudioDevice, AudioData.data(), (uint32_t)AudioData.size()) != 0)
        throw runtime_error("SDL_QueueAudio failed: " + string(SDL_GetError()));

    if (Paused && SDL_GetQueuedAudioSize(AudioDevice) >= MinQueuedAudioSize)
    {
        SDL_PauseAudioDevice(AudioDevice, 0);
        Paused = false;
    }
}
