// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_AudioTx.cpp *#*#*#*#*#*#*#*#*# (C) 2025 DekTec
//
// Example: AvFifo_AudioTx  SMPTE 2110-30 Audio Transmitter
// 
// This example shows how to transmit SMPTE 2110-30 PCM audio packets using the AvFifo
// API. It generates a multichannel sine wave and transmits it over the network.
// 
// Configuration is done through a set of constants below.
// To adapt the example to your setup, update the constants and recompile.
// 
// Note:
//   This example focuses on transmitting SMPTE 2110-30 PCM audio data.
//   For a more complete overview of AvFifo usageincluding device attachment,
//   FIFO lifecycle (start/stop/detach), error handling, and configuration patterns
//   refer to the AvFifo_VideoTx example.
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include required headers -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <chrono>                   // Standard library: timing utilities.
#include <cmath>                    // Standard library: math functions.
#include <iostream>                 // Standard library: I/O streams.
#include <thread>                   // Standard library: threading support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Device selection.
constexpr int CfgDeviceType{2110};  // Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceNo{0};       // Device number: 0 = first matching device in system.

// IP configuration for outgoing 2110-30 audio stream.
const AvFifo::IpPars AvFifo_AudioTx_IpPars
{
    .IpAddr{239, 30, 0, 1},
    .IpVersion{AvFifo::IpProtocolVersion::IPv4},
    .Port{50030},
    .RtpPayloadType{96},
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}  // Must be RTP for 2110-30.
};

// Audio stream transmission configuration.
const AvFifo::St2110::TxConfigAudio AvFifo_AudioTx_Config
{
    .Format{AvFifo::St2110::AudioFormat::L24BE},
    .NumChannels{2},
    .NumSamplesPerIpPacket{48},
    .SampleRate{48000}
};

// The frequency of the generated sine wave.
constexpr int SineFrequency{440};

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Simple multichannel sine wave generator.
struct SineGenerator
{
    SineGenerator();
    void Generate(uint8_t *Samples, int NumSamples);

    int NumChannels{2};
    int Frequency{440};
    int SampleRate{48000};
    int BitDepth{16};
    int SampleNumber{0};
};

// Forward declarations for top-down function presentation.
void AvFifo_AudioTx();
void AudioTxLoop(DtDevice&);
DtTimeOfDay CalculateAudioTimeStamp(DtTimeOfDay StartToD, uint64_t SampleNumber,
                                                                          int SampleRate);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function prints application details and invokes the AvFifo_AudioTx() function.
// Any exceptions are caught and reported.
//
int main()
{
    cout << "\n"
         << "--------------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_AudioTx - SMPTE 2110-30 audio generator\n"
         << "Version " << Version << "\n"
         << "--------------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << " device number " << CfgDeviceNo << "\n"
         << "Destination IP address: "
         << (int)AvFifo_AudioTx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_AudioTx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_AudioTx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_AudioTx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_AudioTx_IpPars.Port << "\n"
         << "Audio format          : ";
    if (AvFifo_AudioTx_Config.Format == AvFifo::St2110::AudioFormat::L24BE)
        cout << "24-bit PCM, ";
    else if (AvFifo_AudioTx_Config.Format == AvFifo::St2110::AudioFormat::L16BE)
        cout << "16-bit PCM, ";
    else
        cout << "PCM, ";
    cout << AvFifo_AudioTx_Config.NumChannels << "-channel, "
         << AvFifo_AudioTx_Config.SampleRate << "Hz\n"
         << "Press Ctrl+C to stop the application.\n\n";

    try {
        AvFifo_AudioTx();
    }
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e)          {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                   {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_AudioTx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Opens the DekTec device, checks network readiness, runs the main transmit loop,
// and finally detaches from the hardware.
//
void AvFifo_AudioTx()
{
    DtDevice Device{};
    DTAPI_RESULT Res = Device.AttachToType(CfgDeviceType, CfgDeviceNo);
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    Res = Device.IsNetworkCardOperational(1, 
                        AvFifo_AudioTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                        AvFifo_AudioTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    try {
        AudioTxLoop(Device);
    }
    catch (AvFifo::DriverError& e)     {cerr << "Driver error: " << e.what() << '\n';}
    catch (AvFifo::SchedulingError& e) {cerr << e.what() << '\n';
    }

    Device.Detach();
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AudioTxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Core function of the example.
// Initializes and configures the TxFifo for SMPTE 2110-30 audio transmission, and
// continuously transmits alternating Audio packet formats.
// 
void AudioTxLoop(DtDevice& Device)
{
    AvFifo::TxFifo AudioTxFifo{};
    AudioTxFifo.Attach(Device, 1);

    AudioTxFifo.SetIpPars(AvFifo_AudioTx_IpPars);
    AudioTxFifo.Configure(AvFifo_AudioTx_Config);
    AudioTxFifo.SetMaxSize(24);

    // Set the initial Time-of-Day (ToD) slightly in the future (now + 10ms) to ensure
    // there is time to prepare the first frames before their intended Tx timestamp.
    DtTimeOfDay StartToD{};
    Device.GetTimeOfDay(StartToD);
    StartToD += 10'000'000;             // 10,000,000ns = 10ms
    StartToD = AvFifo::Tod2Grid_Audio(StartToD, AvFifo_AudioTx_Config.SampleRate);

    SineGenerator Generator{};

    // Keep track of the current sample number in order to calculate the correct Tx
    // timestamp.
    uint64_t SampleNumber = 0;

    // Start transmission; return false means no data can be sent (e.g. IP misconfigured).
    if (!AudioTxFifo.Start())
    {
        AvFifo::FifoStatus Status{AudioTxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        AudioTxFifo.Detach();
        return;
    }

    // Main loop: keep transmitting until a FIFO failure occurs.
    while (true)
    {
        // Check if the FIFO is still healthy and transmitting.
        AvFifo::FifoStatus Status{AudioTxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // Proceed only if there is space in the FIFO.
        if (AudioTxFifo.GetFifoLoad() < AudioTxFifo.GetMaxSize())
        {
            // Generate audio in batches of 1024 samples.
            static int NumSamples = 1024;
            static int FrameSize =
                            (NumSamples * Generator.NumChannels * Generator.BitDepth) / 8;

            // Calculate ToD for the next batch of audio samples.
            DtTimeOfDay ToD = CalculateAudioTimeStamp(StartToD, SampleNumber,
                                                                    Generator.SampleRate);

            // Allocate a new frame from the memory pool.
            AvFifo::Frame* Frame = AudioTxFifo.GetFrameFromMemPool(FrameSize);
            Frame->ToD = ToD;
            Frame->RtpTime = AvFifo::St2110::Tod2Rtp_Audio(ToD, Generator.SampleRate);

            // Generate audio samples.
            Generator.Generate(Frame->Data(), NumSamples);
            Frame->NumValidBytes = FrameSize;

            // Transmit the frame.
            AudioTxFifo.Write(Frame);

            // Advance SampleNumber.
            SampleNumber += NumSamples;
        } else {
            // FIFO full; sleep briefly to reduce CPU usage.
            this_thread::sleep_for(10ms);
        }
    }

    // Clean up the FIFO and detach from hardware.
    AudioTxFifo.Stop();
    AudioTxFifo.Detach();
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.- SineGenerator::SineGenerator -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
SineGenerator::SineGenerator()
{
    if (AvFifo_AudioTx_Config.Format == AvFifo::St2110::AudioFormat::L16BE)
        BitDepth = 16;
    else if (AvFifo_AudioTx_Config.Format == AvFifo::St2110::AudioFormat::L24BE)
        BitDepth = 24;
    NumChannels = AvFifo_AudioTx_Config.NumChannels;
    Frequency = SineFrequency;
    SampleRate = AvFifo_AudioTx_Config.SampleRate;
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SineGenerator::Generate -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void SineGenerator::Generate(uint8_t *Samples, int NumSamples)
{
    static constexpr double PI{3.141592653589793238};
    static constexpr int MaxValue{0x7FFFFFFF};

    for (int SampleIdx = 0; SampleIdx < NumSamples; SampleIdx++)
    {
        int32_t Sample{(int32_t)(MaxValue *
                            sin(2 * PI * Frequency * SampleNumber / (double)SampleRate))};

        for (int ChannelIdx = 0; ChannelIdx < NumChannels; ChannelIdx++)
        {
            if (BitDepth == 16)
            {
                Samples[0] = ((uint32_t)Sample >> 24) & 0xFF;
                Samples[1] = ((uint32_t)Sample >> 16) & 0xFF;
                Samples += 2;
            }
            else if (BitDepth == 24)
            {
                Samples[0] = ((uint32_t)Sample >> 24) & 0xFF;
                Samples[1] = ((uint32_t)Sample >> 16) & 0xFF;
                Samples[2] = ((uint32_t)Sample >>  8) & 0xFF;
                Samples += 3;
            }
        }

        SampleNumber = (SampleNumber + 1) % SampleRate;
    }
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- CalculateAudioTimeStamp -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
DtTimeOfDay CalculateAudioTimeStamp(DtTimeOfDay StartToD, uint64_t SampleNumber,
                                                                           int SampleRate)
{
    DtTimeOfDay ToD = StartToD;

    // Calculate ToD.
    ToD += (SampleNumber / SampleRate) * 1000000000;
    ToD += ((SampleNumber % SampleRate) * 1000000000 + (SampleRate / 2)) / SampleRate;

    // Align ToD to the audio media grid.
    ToD = AvFifo::Tod2Grid_Audio(ToD, SampleRate);

    return ToD;
}
