// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_FastMetadataRx.cpp *#*#*#*#*#*#*# (C) 2025 DekTec
//
// Example: AvFifo_FastMetadataRx  SMPTE 2110-41 Fast Metadata Receiver
//
// This example shows how to receive SMPTE 2110-41 FastMetadata packets using the
// AvFifo API. It connects to a fixed IP and port, receives FastMetadata items, and
// prints their metadata.
//
// Configuration is done through a set of constants below.
// To adapt the example to your setup, update the constants and recompile.
//
// Note:
//   This example focuses on receiving SMPTE 2110-41 FastMetadata items.
//   For a more complete overview of AvFifo usageincluding device attachment,
//   FIFO lifecycle (start/stop/detach), error handling, and configuration patterns
//   refer to the AvFifo_VideoRx example.
// 
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include required headers -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <chrono>                   // Standard library: timing utilities.
#include <iostream>                 // Standard library: I/O streams.
#include <thread>                   // Standard library: threading support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Device selection.
constexpr int CfgDeviceType{2110};  // Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceNo{0};       // Device number: 0 = first matching device in system.

// IP configuration for the incoming 2110-41 FastMetadata stream.
const AvFifo::IpPars AvFifo_FastMetadataRx_IpPars
{
    .IpAddr{239, 41, 0, 1},
    .IpVersion{AvFifo::IpProtocolVersion::IPv4},
    .Port{50041},
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}  // Must be RTP for 2110-41.
};

// FastMetadata stream reception configuration.
//
// This configuration controls which Fast Metadata Items are received from the stream.
// SMPTE 2110-41 defines two kinds of FastMetadata items:
//   - Regular (non-segmented) items
//   - Segmented items, which are transmitted as Segmented Object Streams (SOS)
//
// - Filter:
//     + List of FastMetadata Type values to accept from non-segmented items.
//     + If empty, no non-segmented items are received.
//     + Type is an 8-bit application-defined field in each metadata item.
//
// - SegFilter:
//     + List of Type values to reassemble and receive from segmented items (SOS).
//     + Reassembly is handled internally based on RTP timestamp and continuity.
//
// To receive all metadata types: leave both Filter and SegFilter empty.
//
// Example here:
//   - Accept Type 1 and 2 from regular (non-segmented) metadata
//   - Accept Type 5 from segmented streams
//
const AvFifo::St2110::RxConfigFastMetadata AvFifo_FastMetadataRx_Config
{
    .Filter{1, 2},                  // Regular items: accept types 1 and 2
    .SegFilter{5}                   // Segmented items: accept type 5.
};

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Forward declarations for top-down function presentation.
void AvFifo_FastMetadataRx();
void FastMetadataRxLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function prints application details and invokes the AvFifo_FastMetadataRx()
// function. Any exceptions are caught and reported.
//
int main()
{
    cout << "\n"
         << "-----------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_FastMetadataRx - SMPTE 2110-41 receiver\n"
         << "Version " << Version << "\n"
         << "-----------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << " device number " << CfgDeviceNo << "\n"
         << "Expected IP address: "
         << (int)AvFifo_FastMetadataRx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_FastMetadataRx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_FastMetadataRx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_FastMetadataRx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_FastMetadataRx_IpPars.Port << "\n"
         << "Close this window to stop the application.\n\n";

    try {
        AvFifo_FastMetadataRx();
    }
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e)          {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                   {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_FastMetadataRx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Opens the DekTec device, checks network readiness, runs the main FastMetadata receive
// loop, and finally detaches from the hardware.
//
void AvFifo_FastMetadataRx()
{
    DtDevice Device{};
    DTAPI_RESULT Res = Device.AttachToType(CfgDeviceType, CfgDeviceNo);
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    Res = Device.IsNetworkCardOperational(1, 
                 AvFifo_FastMetadataRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                 AvFifo_FastMetadataRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    try {
        FastMetadataRxLoop(Device);
    }
    catch (AvFifo::DriverError& e) {cerr << "Driver error: " << e.what() << '\n';}

    Device.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- FastMetadataRxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Core function of the example.
// Attaches a FastMetadata RxFifo, configures it, and prints all received items.
//
void FastMetadataRxLoop(DtDevice& Device)
{
    AvFifo::St2110::RxFifoFastMetadata RxFifo{};
    RxFifo.Attach(Device, 1);

    RxFifo.SetIpPars(AvFifo_FastMetadataRx_IpPars);
    RxFifo.Configure(AvFifo_FastMetadataRx_Config);

    if (!RxFifo.Start())
    {
        AvFifo::FifoStatus Status{RxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        RxFifo.Detach();
        return;
    }

    while (true)
    {
        AvFifo::FifoStatus Status{RxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        if (RxFifo.GetFifoLoad() > 0)
        {
            AvFifo::St2110::FastMetadataFrame* Frame{RxFifo.Read()};

            cout << "--------------------------------------------\n";
            cout << "Received FastMetadata: #Items: " << Frame->Items.size() << '\n';
            cout << "--------------------------------------------\n";

            for (AvFifo::St2110::FastMetadataItem Item: Frame->Items)
            {
                cout << "Type:" << Item.Type << '\n';
                if (Frame->IsSegmented)
                    cout << "Segmented frame merged\n";
                else
                    cout << "KBit: " << Item.KBit << '\n';
                cout << "#data words:" << Item.Data.size() << '\n';
                cout << "--------------------------------------------\n";

            }
            RxFifo.ReturnToMemPool(Frame);
        } else {
            this_thread::sleep_for(10ms);
        }
    }

    // Stop transmission and release hardware resources.
    RxFifo.Stop();
    RxFifo.Detach();
}
