// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_FastMetadataTx.cpp *#*#*#*#*#*#*# (C) 2025 DekTec
//
// Example: AvFifo_FastMetadataTx  SMPTE 2110-41 Fast Metadata Transmitter
//
// This example shows how to transmit SMPTE 2110-41 FastMetadata packets using the
// AvFifo API. It periodically sends test FastMetadata frames over the network.
//
// The transmitted metadata is purely synthetic and consists of simple counter values.
// It does not follow any real-world FastMetadata schema or standard format.
// 
// Configuration is done through a set of constants below.
// To adapt the example to your setup, update the constants and recompile.
//
// Note:
//   This example focuses on transmitting FastMetadata frames with synthetic test data.
//   For a more complete overview of AvFifo usageincluding device attachment,
//   FIFO lifecycle (start/stop/detach), error handling, and configuration patterns
//   refer to the AvFifo_VideoTx example.
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include required headers -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <chrono>                   // Standard library: timing utilities.
#include <iostream>                 // Standard library: I/O streams.
#include <thread>                   // Standard library: threading support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Device selection.
constexpr int CfgDeviceType{2110};  // Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceNo{0};       // Device number: 0 = first matching device in system.

// IP configuration for the outgoing 2110-41 FastMetadata stream.
const AvFifo::IpPars AvFifo_FastMetadataTx_IpPars
{
    .IpAddr{239, 41, 0, 1},
    .IpVersion{AvFifo::IpProtocolVersion::IPv4},
    .Port{50041},
    .RtpPayloadType{101},
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}  // Must be RTP for 2110-41.
};

// FastMetadata stream transmission configuration.
// Currently, TxConfigFastMetadata has no configurable parameters.
const AvFifo::St2110::TxConfigFastMetadata AvFifo_FastMetadataTx_Config
{
};

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Forward declarations for top-down function presentation.
void AvFifo_FastMetadataTx();
void FastMetadataTxLoop(DtDevice&);
void GenerateFastMetadataItem(AvFifo::St2110::FastMetadataFrame*, bool);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function prints application details and invokes the AvFifo_FastMetadataTx()
// function. Any exceptions are caught and reported.
//
int main()
{
    cout << "\n"
         << "-----------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_FastMetadataTx - SMPTE 2110-41 transmitter\n"
         << "Version " << Version << "\n"
         << "-----------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << " device number " << CfgDeviceNo << "\n"
         << "Destination IP address: "
         << (int)AvFifo_FastMetadataTx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_FastMetadataTx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_FastMetadataTx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_FastMetadataTx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_FastMetadataTx_IpPars.Port << "\n"
         << "Press Ctrl+C to stop the application.\n\n";

    try {
        AvFifo_FastMetadataTx();
    }
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e)          {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                   {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_FastMetadataTx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Opens the DekTec device, checks network readiness, runs the main FastMetadata transmit
// loop, and finally detaches from the hardware.
//
void AvFifo_FastMetadataTx()
{
    DtDevice Device{};
    DTAPI_RESULT Res = Device.AttachToType(CfgDeviceType, CfgDeviceNo);
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    Res = Device.IsNetworkCardOperational(1, 
                 AvFifo_FastMetadataTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                 AvFifo_FastMetadataTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    try {
        FastMetadataTxLoop(Device);
    }
    catch (AvFifo::DriverError& e)     {cerr << "Driver error: " << e.what() << '\n';}
    catch (AvFifo::SchedulingError& e) {cerr << e.what() << '\n';}

    Device.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- FastMetadataTxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Core function of the example.
// Demonstrates how to transmit SMPTE 2110-41 FastMetadata using the AvFifo API.
// 
// This function alternates between two types of synthetic FastMetadata frames:
//   - Frame 1: Two regular (non-segmented) items:
//       + Type 1: 50 incrementing values (1 to 50)
//       + Type 2: 25 decrementing values (25 to 1)
//   - Frame 2: One segmented item:
//       + Type 5: 600 incrementing values (0 to 599), sent as a segmented object stream
//
// Noteworthy aspects:
//  + The FastMetadata content is synthetic and not based on any real metadata standard.
//  + Frames are aligned to the 50-Hz media clock grid to simulate video-synchronized 
//    timing.
//  + Transmission timing is derived from Time-of-Day (ToD), not an external clock.
//  + AvFifo::Tod2Grid_Video() and AvFifo::Tod2Rtp_Video() are used to align ToD to
//    RTP timestamps.
//
// The loop continues until Ctrl+C is pressed or an error occurs.
//
void FastMetadataTxLoop(DtDevice& Device)
{
    AvFifo::St2110::TxFifoFastMetadata TxFifo{};
    TxFifo.Attach(Device, 1);

    TxFifo.SetIpPars(AvFifo_FastMetadataTx_IpPars);
    TxFifo.Configure(AvFifo_FastMetadataTx_Config);

    // Set the initial Time-of-Day (ToD) slightly in the future (now + 100ms) to ensure
    // there is time to prepare the first frames before their intended Tx timestamp.
    DtTimeOfDay ToD{};
    Device.GetTimeOfDay(ToD);
    ToD += 100'000'000;             // 100,000,000ns = 100ms

    if (!TxFifo.Start())
    {
        AvFifo::FifoStatus Status{TxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        TxFifo.Detach();
        return;
    }

    bool Segmented = false;
    while (true)
    {
        AvFifo::FifoStatus Status{TxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // Proceed only if the FIFO has space for another frame.
        if (TxFifo.GetFifoLoad() < TxFifo.GetMaxSize())
        {
            // Align frame transmission to the video media grid. 
            ToD = AvFifo::Tod2Grid_Video(ToD, {50, 1});

		    // Allocate and populate a new FastMetadataFrame from the memory pool.
            AvFifo::St2110::FastMetadataFrame* Frame = TxFifo.GetFrameFromMemPool();

            // Fill in the timestamps (ToD for scheduling, RTP time for packet headers).
            Frame->ToD = ToD;
            Frame->RtpTime = AvFifo::St2110::Tod2Rtp_Video(ToD);
            
            // Fill the frame with test metadata content.
            GenerateFastMetadataItem(Frame, Segmented);
            
            // Submit frame for transmission.
            TxFifo.Write(Frame);

            // Alternate frame content between segmented and non-segmented format.
            Segmented = !Segmented;

		    // Advance ToD to the next frame time, which is 20ms later (50Hz).
            ToD += 20'000'000;      // 20,000,000ns = 20ms
        } else {
            // FIFO full  sleep briefly before retrying to avoid busy-waiting.
            this_thread::sleep_for(10ms);
        }
    }

    // Stop transmission and release hardware resources.
    TxFifo.Stop();
    TxFifo.Detach();
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- GenerateFastMetadataItem -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void GenerateFastMetadataItem(AvFifo::St2110::FastMetadataFrame* Frame, bool Segmented)
{
    Frame->IsSegmented = Segmented;

    if (Segmented)
    {
        // Create a single segmented item of type 5.
        AvFifo::St2110::FastMetadataItem Item;
        Item.Type = 5;
        Item.KBit = false;          // Kbit not used for segmented items.

        for (int i=0; i<600; i++)
            Item.Data.push_back(i);
        Frame->Items.push_back(Item);

    } else {
        // Create two regular (non-segmented) items.

        // First item: Type 1, incrementing values 1 to 50.
        AvFifo::St2110::FastMetadataItem Item;
        Item.Type = 1;
        Item.KBit = false;          // Optional use. Always false in this test example.

        for (int i=1; i<=50; i++)
            Item.Data.push_back(i);
        Frame->Items.push_back(Item);

        // Second item: Type 2, decrementing values 25 to 1.
        Item.Type = 2;
        Item.Data.clear();

        for (int i=25; i>=1; i--)
            Item.Data.push_back(i);
        Frame->Items.push_back(Item);
    }
}
