// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_PktTiming.cpp *#*#*#*#*#*#*#*#* (C) 2025 DekTec
//
// Example: AvFifo_PktTiming - Access arrival timestamps of SMPTE 2110 IP packets
//
// This example demonstrates how to access the arrival timestamps of SMPTE 2110 packets
// and compute inter-packet arrival times in nanoseconds by subtracting timestamps of
// successive packets. The computed values are printed to the console.
//
// In a real-world application, arrival timestamps could be stored in a database,
// analyzed for jitter, and used to compute statistics for logging or graphical display.
//
// Configuration is done through a set of constants below.
// To adapt the example to your setup, update the constants and recompile.
// 
// Notes:
//  - Printing inter-packet times may cause packet drops due to the overhead of std::cout.
//  - For a more detailed explanation of the boilerplate code, including initialization,
//    exception handling, error handling, and cleanup, refer to the comments in the
//    AvFifo_VideoRx and AvFifo_VideoTx examples.
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include required headers -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <chrono>                   // Standard library: date and time utilities.
#include <iostream>                 // Standard library: input/output operations.
#include <stdexcept>                // Standard library: std::runtime_error.
#include <thread>                   // Standard library: Thread support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Device selection.
constexpr int CfgDeviceType{2110};  // Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceNo{0};       // Device number: 0 = first matching device in system.

// Configuration for the incoming IP stream.
const AvFifo::IpPars AvFifo_PktTiming_IpPars
{
    .IpAddr{239, 10, 10, 10},
    .IpVersion{AvFifo::IpProtocolVersion::IPv4},
    .Port{5000},
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}
};

// Configure for receiving raw SMPTE 2110 packets including timestamps.
// This example only processes arrival timestamps. If IncludeRtpHeader is set to true,
// the application could be extended to access the RTP header and analyze RTP timestamps.
const AvFifo::St2110::RxConfigRaw AvFifo_PktTiming_Config
{
    .IncludeRtpHeader{false},
    .MaxRate{1500000000}
};

// Number of packets for which inter-packet time is logged.
constexpr int MaxPacketCount{101};

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Forward declarations to enable top-down function presentation.
void AvFifo_PktTiming();
void PktTimingLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int argc, char* argv[])
{
    cout << "\n"
         << "------------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_PktTiming - Access arrival timestamps\n"
         << "Version " << Version << "\n"
         << "------------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << " device number: " << CfgDeviceNo << "\n"
         << "Expected IP address   : "
         << (int)AvFifo_PktTiming_IpPars.IpAddr[0] << "."
         << (int)AvFifo_PktTiming_IpPars.IpAddr[1] << "."
         << (int)AvFifo_PktTiming_IpPars.IpAddr[2] << "."
         << (int)AvFifo_PktTiming_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_PktTiming_IpPars.Port << "\n";

    try {
        AvFifo_PktTiming();
    }
    catch (AvFifo::UsageError& e) {cerr << "Usage error: " << e.what() << '\n';}
    catch (exception& e)          {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                   {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_PktTiming -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Attaches to the hardware, runs the main loop, and performs cleanup.
//
void AvFifo_PktTiming()
{
    DtDevice Device{};
    DTAPI_RESULT Res{Device.AttachToType(CfgDeviceType, CfgDeviceNo)};
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    Res = Device.IsNetworkCardOperational(1, 
                      AvFifo_PktTiming_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                      AvFifo_PktTiming_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    try {
        PktTimingLoop(Device);
    }
    catch (AvFifo::DriverError& e) {cerr << "Driver error: " << e.what() << '\n';}
    catch (AvFifo::HardwarePipeUnavailable& e) {cerr << e.what() << '\n';}

    Device.Detach();
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- PktTimingLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// This function is the core of the example program.
// It receives raw RTP packets and logs the inter-packet arrival time in nanoseconds.
//
void PktTimingLoop(DtDevice& Device)
{
    // Initialize an RxFifo object and associate it with the first (and only) network 
    // port of the device. We force a hardware pipe to be used here (ForceHwPipe).
    // While not strictly necessary for accurate timestamps, a hardware pipe reduces 
    // CPU load with high-bitrate SMPTE 2110 streams.
    AvFifo::RxFifo PktTimingFifo{};
    PktTimingFifo.Attach(Device, 1, AvFifo::HwOrSwPipe::ForceHwPipe);

    PktTimingFifo.SetIpPars(AvFifo_PktTiming_IpPars);
    PktTimingFifo.Configure(AvFifo_PktTiming_Config);

    // In high-throughput scenarios, this may need tuning.
    PktTimingFifo.SetMaxSize(1000);

    if (!PktTimingFifo.Start())
    {
        AvFifo::FifoStatus Status{PktTimingFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';

        PktTimingFifo.Detach();
        return;
    }

    DtTimeOfDay PrevToD{};
    for (int PacketCount{0}; PacketCount<MaxPacketCount;)
    {
        // Check the current status of the FIFO.
        AvFifo::FifoStatus Status{PktTimingFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // If at least one packet is available, retrieve and process it.
        if (PktTimingFifo.GetFifoLoad() > 0)
        {
            AvFifo::Frame* Frm{PktTimingFifo.Read()};

            // If this is not the first packet, compute delta against previous timestamp.
            if (PacketCount > 0)
            {
                int64_t PacketDelta{Frm->ToD - PrevToD};
                cout << "Packet delta: " << PacketDelta << "ns\n";
            }
            PrevToD = Frm->ToD;
            PacketCount++;
            
            // Return the frame to the memory pool for reuse.
            PktTimingFifo.ReturnToMemPool(Frm);
        } else {
            // If no packet is available, sleep briefly to avoid busy-waiting.
            std::this_thread::sleep_for(1ms);
        }
    }

    // Stop receiving frames and detach the RxFifo object to clean up.
    PktTimingFifo.Stop();
    PktTimingFifo.Detach();
}
