// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_VideoRx.cpp *#*#*#*#*#*#* (C) 2023-2025 DekTec
//
// Example: AvFifo_VideoRx - SMPTE 2110-20 video receiver
//
// This code serves as a practical code example of receiving a SMPTE 2110 video stream
// and displaying it in a window using the SDL library with video scaling by 0.5.
// 
// The example covers key concepts such as device configuration, receive FIFO management, 
// processing of received frames, SDL event handling, resource cleanup, and exception 
// handling.
// 
// Notes:
//  - This example uses a fixed IP address, port number, video format, etc.
//    These configuration parameters are defined in the 'Configuration' section below.
//  - The application downscales the video by a factor of 2, reducing the size of the 
//    video content to fit into the display window.
// 
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// -.-.-.-.-.-.-.-.-.-.-.- Include necessary headers and libraries -.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // AvFifo API for SMPTE 2110; also includes DTAPI.h.
#include <SDL.h>                    // Cross-platform library providing low-level access 
                                    // to video, audio and other I/O functions.
#include <chrono>                   // For duration literals like ms.
#include <iostream>                 // Standard library: input/output operations.
#include <thread>                   // For std::this_thread::sleep_for.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// +=+=+=+=+=+=+=+=+=+=+=+=+=+ Start of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=

// If a config value can be changed without changing the application code, it is 
// labeled "CONFIGURABLE".

// Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceType{2110};               // CONFIGURABLE

// Device number: 0 = first device in system, 1 = second one, etc.
constexpr int CfgDeviceNo{0};                    // CONFIGURABLE

// IP receive configuration parameters.
// - IpAddr           : Source IP address. Can be unicast or multicast.
// - IpVersion        : Use IPv4 or IPv6.
// - Port             : UDP port number to listen on. Must match the transmitters port.
// - TransportProtocol: Must be set to RTP for SMPTE 2110-20 reception.
//                      Note: The receiver does not validate the RTP payload type.
const AvFifo::IpPars AvFifo_VideoRx_IpPars
{
    .IpAddr{239, 10, 10, 10},                    // CONFIGURABLE: unicast or multicast
    .IpVersion{AvFifo::IpProtocolVersion::IPv4}, // CONFIGURABLE
    .Port{5000},                                 // CONFIGURABLE
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp}
};

// SMPTE 2110-20 video reception configuration.
// - Format: We use 'Uyvy422_10b_to_8b' to instruct the AvFifo API to convert incoming
//   10-bit UYVY video to 8-bit format automatically. This simplifies downstream
//   processing in the application by avoiding manual unpacking of 10-bit pixels.
//
// Note: To receive video in a different format, this program must be updated to handle
//       the new pixel layout accordingly.
const AvFifo::St2110::RxConfigVideo AvFifo_VideoRx_Config
{
    .Format{AvFifo::St2110::RxFrameFormat::Uyvy422_10b_to_8b}
};

// Fixed incoming frame size (1920x1080).
constexpr int FrameWidth{1920};     // CONFIGURABLE
constexpr int FrameHeight{1080};    // CONFIGURABLE

// The factor by which to scale down the video when displaying in an SDL window.
constexpr int Scale{2};             // CONFIGURABLE

// =+=+=+=+=+=+=+=+=+=+=+=+=+=+ End of Configuration Section +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

// Stores the SDL state and takes care of proper initialization/destruction.
struct SdlContext
{
    SdlContext();                   // Initializes SDL.
    ~SdlContext();                  // Cleans up SDL.
    void Render(AvFifo::Frame*);    // Renders a frame to the Window.
    SDL_Renderer* Renderer{};
    SDL_Texture* Texture{};
    SDL_Window* Window{};
};

// Forward declarations to enable top-down function presentation.
void AvFifo_VideoRx();
void VideoRxLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Entry point of the application.
//
// This function prints a startup banner with configuration details, and then invokes
// the main AvFifo_VideoRx() function. Execution is wrapped in a try-catch block to 
// demonstrate proper exception handling in a production-style application.
//
// Exceptions at this level indicate critical errors that prevent the application from
// continuingfor example, no DTA-2110 device present, invalid IP configuration, or a
// missing driver installation.
//
int main(int argc, char* argv[])
{
    cout << "\n"
         << "-------------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_VideoRx - SMPTE 2110-20 video receiver\n"
         << "Version " << Version << "\n"
         << "-------------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << ": " << CfgDeviceNo << "\n"
         << "Expected IP address   : "
         << (int)AvFifo_VideoRx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_VideoRx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_VideoRx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_VideoRx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_VideoRx_IpPars.Port << "\n"
         << "Expected video format : 10-bit packed, 4:2:2 UYVY chroma subsampling, "
         << FrameWidth << "x" << FrameHeight << "\n"
         << "Scale-down factor     : " << Scale << "\n\n"
         << "Closing the Video Window will quit the application.\n";

    // Run the main application logic.
    // This block handles all top-level exceptions that are not caught inside
    // AvFifo_VideoRx().
    try {
        AvFifo_VideoRx();
    }

    // AvFifo::UsageError is thrown by the AvFifo API for incorrect usage, such as
    // invalid configuration or calling functions in an invalid state.
    // These errors typically indicate a programming mistake in the application.
    catch (AvFifo::UsageError& e)
    {
        cerr << "Usage error: " << e.what() << '\n';
    }

    // AvFifo::DriverError indicates a failure reported by the underlying driver layer.
    // This exception is typically thrown when the network interface or device
    // is not in a usable state. Common causes include:
    //   - The network interface is unplugged or has no link
    //   - No IP address is configured
    //   - Required drivers (e.g., for VLAN or general network support) are missing
    //   - The DekTec PCIe driver is not functioning properly
    // These conditions are detected internally by the AvFifo library when starting
    // or checking a TxFifo or RxFifo.
    catch (AvFifo::DriverError& e)
    {
        cerr << "Driver error: " << e.what() << '\n';
    }

    // Catch any standard exceptions (e.g. std::bad_alloc, std::runtime_error).
    // This block handles general runtime failures.
    catch (exception& e)
    {
        cerr << "Exception: " << e.what() << '\n';
    }

    // Fallback for unknown exception types.
    // This ensures that even unexpected faults do not crash the program ungracefully.
    catch (...)
    {
        cerr << "Caught an unknown exception\n";
    }

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_VideoRx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Opens the specified DekTec device, checks that the network interface is ready,
// and then invokes the main video receive loop. After reception stops (e.g., due to
// an error or end of input), the function detaches cleanly from the hardware.
//
// This routine demonstrates proper resource handling: device attachment, validation,
// usage, and detachment. It also shows how to catch and report hardware- and
// driver-level errors.
//
void AvFifo_VideoRx()
{
    // Create a DtDevice object and attach it to the specified DekTec device.
    DtDevice Device{};
    DTAPI_RESULT Res{Device.AttachToType(CfgDeviceType, CfgDeviceNo)};
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails: " + string(DtapiResult2Str(Res)));

    // Verify that the network interface (port 1) is operational.
    // This is particularly relevant for applications that auto-launch on boot,
    // where the OS might not yet have initialized the NIC.
    Res = Device.IsNetworkCardOperational(1, 
                        AvFifo_VideoRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                        AvFifo_VideoRx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    // The NIC is ready; proceed to the main video reception loop.
    // Errors from AvFifo are caught and reported below.
    try {
        VideoRxLoop(Device);
    }

    // AvFifo::HardwarePipeUnavailable is thrown when no Rx hardware pipes are free.
    // Hardware pipes are a limited resource and may be in use by other applications.
    catch (AvFifo::HardwarePipeUnavailable& e)
    {
        cerr << e.what() << '\n';
    }

    // Cleanup: Detach the Device object from the hardware.
    Device.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- VideoRxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Core receive loop for the AvFifo_VideoRx example.
// Handles configuration of the AvFifo RxFifo, starts video reception, and renders the
// received frames using SDL.
//
// This function demonstrates best practices for using the AvFifo API, including:
//  - Attaching and configuring a receive FIFO.
//  - Graceful handling of startup failures and runtime errors.
//  - Integration with a video rendering library (SDL).
//
void VideoRxLoop(DtDevice& Device)
{
    // Create an RxFifo and attach it to port 1 of the SmartNIC.
    // A hardware pipe is explicitly requested for optimal performance.
    // If no hardware pipes are available (e.g. all in use), an exception will be thrown
    // later by RxFifo::Start(), not here.
    AvFifo::RxFifo VideoRxFifo{};
    VideoRxFifo.Attach(Device, 1, AvFifo::HwOrSwPipe::ForceHwPipe);

    // Configure the RxFifo with IP parameters and the expected SMPTE 2110-20 format.
    VideoRxFifo.SetIpPars(AvFifo_VideoRx_IpPars);
    VideoRxFifo.Configure(AvFifo_VideoRx_Config);

    // Initialize SDL for displaying the received video frames.
    SdlContext Sdl{};

    // Attempt to start receiving.
    // Two failure types can occur:
    //  - Exception: Serious issue like a driver failure or no free hardware pipes.
    //  - Error return: Operational condition such as a disconnected network cable.
    if (!VideoRxFifo.Start())
    {
        // If VideoRxFifo fails to start, retrieve the RxFifo error using GetStatus().
        AvFifo::FifoStatus Status{VideoRxFifo.GetStatus()};

        // Generate a human-readable error message for clarity and problem-solving.
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';

        // To prevent potential resource leakage, detach the VideoRxFifo.
        VideoRxFifo.Detach();

        // Halt further execution due to the failure in video frame reception.
        return;
    }

    // Main loop: continue until a quit event is detected.
    // In this example, the QuitFlag is only set by an SDL quit event.
    bool QuitFlag{false};
    while (!QuitFlag)
    {
        // Monitor FIFO health.
        AvFifo::FifoStatus Status{VideoRxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            break;
        }

        // Poll for SDL events. If an SDL 'quit' event is detected (closing the window), 
        // break out of the loop.
        SDL_Event SdlEvent;
        while (SDL_PollEvent(&SdlEvent))
            if (SdlEvent.type == SDL_QUIT)
                QuitFlag = true;

        if (QuitFlag)
            break;
        
        // If a video frame is available, render it in the SDL Window and return the
        // frame to the memory pool.
        if (VideoRxFifo.GetFifoLoad() > 0)
        {
            AvFifo::Frame* Frm{VideoRxFifo.Read()};
            Sdl.Render(Frm);
            VideoRxFifo.ReturnToMemPool(Frm);
        } else {
            // No frame available: sleep briefly to avoid busy-waiting.
            this_thread::sleep_for(10ms);
        }
    }

    // Cleanup: stop streaming and detach from hardware.
    VideoRxFifo.Stop();
    VideoRxFifo.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::RenderFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Renders an AvFifo frame to an SDL window.
//
void SdlContext::Render(AvFifo::Frame* Frm)
{
    SDL_UpdateTexture(Texture, nullptr, Frm->Data(), FrameWidth*2);
    SDL_RenderCopy(Renderer, Texture, nullptr, nullptr);
    SDL_RenderPresent(Renderer);
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::SdlContext -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Initializes SDL for video rendering, and create a window, texture, and renderer for
// displaying video in a window.
//
SdlContext::SdlContext()
{
    if (SDL_Init(SDL_INIT_VIDEO) < 0)
        throw runtime_error("SDL initialization failed: " + string(SDL_GetError()));

    Window = SDL_CreateWindow("AvFifo_VideoRx - Video Window",
                               SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 
                               FrameWidth/Scale, FrameHeight/Scale, SDL_WINDOW_RESIZABLE);
    if (!Window)
        throw runtime_error("SDL_CreateWindow failed: " + string(SDL_GetError()));

    Renderer = SDL_CreateRenderer(Window, -1, 0);
    if (!Renderer)
        throw runtime_error("SDL_CreateRenderer failed: " + string(SDL_GetError()));

    // The texture uses the UYVY pixel format.
    Texture = SDL_CreateTexture(Renderer, SDL_PIXELFORMAT_UYVY,
                                    SDL_TEXTUREACCESS_STREAMING, FrameWidth, FrameHeight);
    if (!Texture)
        throw runtime_error("SDL_CreateTexture failed: " + string(SDL_GetError()));
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- SdlContext::~SdlContext -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Destroys the SDL resources. 
//
SdlContext::~SdlContext()
{
    if (Texture)
    {
        SDL_DestroyTexture(Texture); 
        Texture = nullptr;
    }
    if (Renderer) 
    {
        SDL_DestroyRenderer(Renderer); 
        Renderer = nullptr;
    }
    if (Window) 
    {
        SDL_DestroyWindow(Window);
        Window = nullptr;
    }
}
