// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* AvFifo_VideoTx.cpp *#*#*#*#*#*#* (C) 2023-2025 DekTec
//
// Example: AvFifo_VideoTx - SMPTE 2110-20 video transmitter
// 
// This example demonstrates how to transmit SMPTE 2110-20 uncompressed video with
// the AvFifo API. It generates a synthetic test pattern: animated balls bouncing around 
// the screen, and sends it in real time over IP as a live video stream.
//
// Configuration parameters - such as IP address, port, and video format - are defined
// in 'Configuration.h'. To adapt the example to your environment, adjust the constants
// in that file and recompile.
// 
// Note:
//   This example contains limited in-line commentary. For a more thorough guide to
//   AvFifo setup, transmission lifecycle, and error handling, refer to the companion 
//   example: AvFifo_VideoRx.
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include required headers -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "Configuration.h"          // Constants for configuring this specific example.
#include "DTAPI_AvFifo.h"           // DekTec's AvFifo API, part of DTAPI.
#include "Generator.h"              // SMPTE 2110 video frame generator.
#include <chrono>                   // Standard library: date and time utilities.
#include <iostream>                 // Standard library: Console input/output operations.
#include <thread>                   // Standard library: Thread support.

const char* Version{"Jun2025.1"};

using namespace std;                // Used for clarity in this example.
using namespace chrono_literals;    // Enables time literals like 10ms.

// Forward declarations to enable top-down function presentation.
void AvFifo_VideoTx();
void VideoTxLoop(DtDevice&);

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// The main function initiates the execution of the program and manages fatal exceptions.
//
int main(int argc, char* argv[])
{
    cout << "\n"
         << "--------------------------------------------------------------------------\n"
         << "DekTec AvFifo API Example - AvFifo_VideoTx - SMPTE 2110-20 video generator\n"
         << "Version " << Version << "\n"
         << "--------------------------------------------------------------------------\n\n"
         << "DTA-" << CfgDeviceType << ": " << CfgDeviceNo << "\n"
         << "Destination IP address: "
         << (int)AvFifo_VideoTx_IpPars.IpAddr[0] << "."
         << (int)AvFifo_VideoTx_IpPars.IpAddr[1] << "."
         << (int)AvFifo_VideoTx_IpPars.IpAddr[2] << "."
         << (int)AvFifo_VideoTx_IpPars.IpAddr[3] << ", "
         << "port " << AvFifo_VideoTx_IpPars.Port << "\n"
         << "Video format          : 10-bit packed, 4:2:2 UYVY chroma subsampling, "
         << FrameWidth << "x" << FrameHeight << "\n"
         << "Moving ball simulation: " << NumMovingBalls << " balls, "
         << "radius " << MinBallRadius << "-" << MaxBallRadius << ", "
         << "initial speed " << MinBallSpeed << "-" << MaxBallSpeed << "pixels/frame\n\n"
         << "Close this window to quit the application.\n";

    try {
        AvFifo_VideoTx();
    }
    catch (AvFifo::UsageError& e)  {cerr << "Usage error: " << e.what() << '\n';}
    catch (AvFifo::DriverError& e) {cerr << "Driver error: " << e.what() << '\n';}
    catch (exception& e)           {cerr << "Exception: " << e.what() << '\n';}
    catch (...)                    {cerr << "Caught an unknown exception\n";}

    // Keep console window open briefly to allow reading of log or error messages.
    this_thread::sleep_for(3000ms);

    return 0;
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- AvFifo_VideoTx -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Opens the DekTec device, checks network readiness, runs the main transmit loop,
// and finally detaches from the hardware.
//
void AvFifo_VideoTx()
{
    DtDevice Device{};
    DTAPI_RESULT Res = Device.AttachToType(CfgDeviceType, CfgDeviceNo);
    if (Res != DTAPI_OK)
        throw runtime_error("Attach fails:" + string(DtapiResult2Str(Res)));

    // Verify that network port 1 is operational.
    Res = Device.IsNetworkCardOperational(1,
                        AvFifo_VideoTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv4,
                        AvFifo_VideoTx_IpPars.IpVersion==AvFifo::IpProtocolVersion::IPv6);
    if (Res != DTAPI_OK)
    {
        Device.Detach();
        throw runtime_error("Networking error: " + string(DtapiResult2Str(Res)));
    }

    // Network card is operational, proceed to invoke the main loop.
    try {
        VideoTxLoop(Device);
    }
    catch (AvFifo::HardwarePipeUnavailable& e) {cerr << e.what() << '\n';}
    catch (AvFifo::SchedulingError& e) {cerr << e.what() << '\n';}

    Device.Detach();
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- VideoTxLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Core transmit loop for the AvFifo_VideoTx example.
// Sets up the AvFifo TxFifo for SMPTE 2110-20 uncompressed video transmission,
// generates synthetic video frames using a simple simulation, and streams them over IP.
//
// Key aspects demonstrated:
//  - Use of a hardware pipe for high-performance video transmission.
//  - Media-clock aligned transmission using ToD and RTP time alignment.
//  - Integration with a user-defined frame generator.
//  - Best practices for start-up validation and clean shutdown.
//
void VideoTxLoop(DtDevice& Device)
{
    // Create a TxFifo for transmitting SMPTE 2110-20 video frames.
    AvFifo::TxFifo VideoTxFifo{};

    // Attach the TxFifo to network port 1 using a hardware pipe.
    VideoTxFifo.Attach(Device, 1, AvFifo::HwOrSwPipe::ForceHwPipe);

    // Configure network parameters (IP address, port, protocol).
    VideoTxFifo.SetIpPars(AvFifo_VideoTx_IpPars);

    // Configure SMPTE 2110-20 stream parameters (e.g. frame format).
    VideoTxFifo.Configure(AvFifo_VideoTx_Config);

    // Calculate size of a single video frame (UYVY 10-bit packed format).
    size_t FrameSize{FrameWidth*FrameHeight*20 / 8};

    // Create and initialize the frame generator.
    Generator FrameGen{};
    FrameGen.Initialize({FrameWidth, FrameHeight});

    // Initialize ToD to current time. 
    // Add 100ms to have time to create the first few frames.
    DtTimeOfDay ToD{};
    Device.GetTimeOfDay(ToD);
    ToD += 100'000'000;             // 100,000,000ns = 100ms

    // Initiates video frame transmission. In case of failure, an exception is thrown 
    // or 'Start()' returns false.
    if (!VideoTxFifo.Start())
    {
        AvFifo::FifoStatus Status{VideoTxFifo.GetStatus()};
        cout << "Failed to start: " << AvFifo::FifoStatusToMessage(Status) << '\n';
        VideoTxFifo.Detach();
        return;
    }

    // Main loop: keep generating and sending frames until a quit event occurs. 
    // In this example, the QuitFlag is only set on failures.
    bool QuitFlag{false};
    while (!QuitFlag)
    {
        // Check FIFO status. On failure, log message and exit loop.
        AvFifo::FifoStatus Status{VideoTxFifo.GetStatus()};
        if (Status != AvFifo::FifoStatus::Ok)
        {
            cout << "Failure: " << AvFifo::FifoStatusToMessage(Status) << '\n';
            QuitFlag = true;
        }

        // Proceed only if FIFO has space for another frame.
        else if (VideoTxFifo.GetFifoLoad() < VideoTxFifo.GetMaxSize())
        {
            // Align transmission time to the 50-Hz video media grid.
            ToD = AvFifo::Tod2Grid_Video(ToD, {50, 1});

		    // Allocate a frame and populate it with synthetic video.
            AvFifo::Frame* Frame{VideoTxFifo.GetFrameFromMemPool(FrameSize)};
            Frame->ToD = ToD;
            Frame->RtpTime = AvFifo::St2110::Tod2Rtp_Video(ToD);
            FrameGen.Generate(Frame->Data());
            Frame->NumValidBytes = (int)FrameSize;

		    // Transmit the frame.
            VideoTxFifo.Write(Frame);

		    // Advance ToD to next video frame time (20ms later for 50Hz).
            ToD += 20'000'000;      // 20,000,000ns = 20ms
        } else {
            // FIFO full  sleep briefly to avoid busy-waiting.
            this_thread::sleep_for(10ms);
        }
    }

    // Cleanup: Stop transmitting frames and release resources.
    VideoTxFifo.Stop();
    VideoTxFifo.Detach();
}
