// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Configuration.h *#*#*#*#*#*#*#*#*#* (C) 2023 DekTec
//
// Code example AvFifo_VideoTx - Configuration
// 
// Standard configuration:
// - Use first DTA-2110 in system
// - Video frame resolution: 1920x1080 pixels
// - IP address for transmission: 239, 10, 10, 10, port 5000
// - SMPTE 2110-20 video format: 10-bit YCbCr-4:2:2, 50Hz
// - Number of balls in the simulation: 200
// - Ball radius: 5-30 pixels, ball speed 0-2 pixels/frame
//
// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// -.-.-.-.-.-.-.-.-.-.-.- Include necessary headers and libraries -.-.-.-.-.-.-.-.-.-.-.-

#include "DTAPI_AvFifo.h"           // DekTec's AvFifo API, part of DTAPI.
#include "Generator.h"              // SMPTE 2110 video frame generator.

using namespace std;                // Used for clarity in this example.

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// If a config value can be changed without changing the application code, it is 
// labeled "CONFIGURABLE".

// Device type: 2110 = DTA-2110, 2125 = DTA-2125
constexpr int CfgDeviceType{2110};               // CONFIGURABLE

// Device number: 0 = first device in system, 1 = second one, etc.
constexpr int CfgDeviceNo{0};                    // CONFIGURABLE

// -.-.-.-.-.-.-.-.-.-.-.-.-.- SMPTE 2110 video configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.

// Fixed frame size (1920x1080 pixels).
constexpr int FrameWidth{1920};                  // CONFIGURABLE
constexpr int FrameHeight{1080};                 // CONFIGURABLE

// IP transmit configuration parameters.
// - IpAddr           : Destination IP address. Can be unicast or multicast, depending on
//                     the receiver setup.
// - IpVersion        : Use IPv4 or IPv6.
// - Port             : UDP port number. Must match the receivers expected port.
// - RtpPayloadType   : RTP payload type in the dynamic range (96127). Not fixed by
//                      SMPTE 2110-20; must match the receiver's expectation.
// - TransportProtocol: Must be RTP for all SMPTE 2110-20 transmissions.
const AvFifo::IpPars AvFifo_VideoTx_IpPars
{
    .IpAddr{239, 10, 10, 10},                    // CONFIGURABLE: unicast or multicast
    .IpVersion{AvFifo::IpProtocolVersion::IPv4}, // CONFIGURABLE
    .Port{5000},                                 // CONFIGURABLE
    .RtpPayloadType{96},                         // CONFIGURABLE: 96-127
    .TransportProtocol{AvFifo::IpTransportProtocol::Rtp} 
};

// SMPTE 2110 video format.
const AvFifo::St2110::TxConfigVideo AvFifo_VideoTx_Config
{
    .Format{AvFifo::St2110::TxFrameFormat::Uyvy422_10b},
    .Packing{false, AvFifo::St2110::PackingMode::General, -1},
    .Resolution{FrameWidth, FrameHeight},
    .Timing{{50, 1}, AvFifo::St2110::Scheduling::Linear, 
            AvFifo::St2110::VideoScanning::Progressive}
};

// .-.-.-.-.-.-.-.-.-.-.-.- Moving balls simulation configuration -.-.-.-.-.-.-.-.-.-.-.-.

// Number of simulated moving balls.
constexpr int NumMovingBalls{200};               // CONFIGURABLE

// Minimum and maximum radius of the balls.
constexpr int MinBallRadius{5};                  // CONFIGURABLE
constexpr int MaxBallRadius{30};                 // CONFIGURABLE

// Minimum and maximum initial velocity of the balls in #pixels moved per frame.
constexpr double MinBallSpeed{0};                // CONFIGURABLE
constexpr double MaxBallSpeed{2};                // CONFIGURABLE

// The frame's background color.
constexpr Color BackgroundColor{80, 200, 0};     // CONFIGURABLE
