// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Generator.cpp *#*#*#*#*#*#*#*#*#*# (C) 2023 DekTec
//
// Code example AvFifo_VideoTx - Moving balls frame generator - Implementation
//
// Note:
// - Although this file is part of the AvFifo_VideoTx code example, it is primarily 
//   concerned with the pattern generation aspect, and does not directly interact with  
//   the AvFifo API.
//
// Refer to 'Generator.h' for a description of the frame generation.
//

// -.-.-.-.-.-.-.-.-.-.-.- Include necessary headers and libraries -.-.-.-.-.-.-.-.-.-.-.-

#include "Configuration.h"          // Constants for configuring this specific example.
#include "Generator.h"              // SMPTE 2110 video frame generator.
#include <iostream>                 // Standard library: Console input/output operations.
#include <random>                   // Standard library: Random number generation.

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Generator::Generate -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Main function of the test generator.
// It renders the moving balls in 24-bit YUV format, and then converts the result to
// 10-bit UYVY. Subsequently, it computes the positions of the balls for the next frame.
//
void Generator::Generate(uint8_t* FrameData)
{
    Img.Clear(BackgroundColor);     // Clear the background with a fixed color.
    for (MovingBall& Ball : Balls)
        Ball.Draw(Img);             // Render each moving ball.
    Img.ConvertToUyvy10(FrameData); // Convert to 10-bit packed UYVY format.
    UpdatePhysics();                // Compute new ball positions and velocities.
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.- FrameGenerator::Initialize -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Initializes the frame generator.
//
void Generator::Initialize(Size BoundingBox)
{
    FrameSize = BoundingBox;
    Img.Init(FrameSize);

    Balls.clear();

    // Initialize the moving balls with random position and velocity.
    random_device RandomDevice;
    mt19937 RndGen(RandomDevice());
    uniform_real_distribution<> PosX(0, FrameSize.w);
    uniform_real_distribution<> PosY(0, FrameSize.h);
    uniform_real_distribution<> Vel(-MaxBallSpeed, MaxBallSpeed);
    uniform_int_distribution<> Col(0, 255);
    uniform_int_distribution<> Rad(MinBallRadius, MaxBallRadius);

    // Lambda for checking whether balls overlap.
    auto AnyOverlap = [&](const MovingBall& NewBall)
    {
        for (const MovingBall& Ball : Balls)
            if (NewBall.DoesOverlap(Ball))
                return true;        // Overlap detected.
        return false;               // No overlap.
    };

    // Generate moving balls one by one.
    for (int i{0}; i<NumMovingBalls; ++i)
    {
        // Find a ball position that does not overlap other billiard balls.
        MovingBall Ball;
        do {
            Ball.p = {PosX(RndGen), PosY(RndGen)};
            Ball.v = {Vel(RndGen), Vel(RndGen)};
            Ball.r = Rad(RndGen);
        } while (AnyOverlap(Ball) || !Ball.FitsInBox(FrameSize));

        // Complete initialization of the ball and prepare for rendering by drawing the
        // ball in a pixel cache.
        Ball.MovingBounds = FrameSize;
        Ball.m = Ball.r * Ball.r;
        Ball.Prepare({(uint8_t)Col(RndGen), (uint8_t)Col(RndGen),(uint8_t) Col(RndGen)});

        // Add ball to the collection of moving balls.
        Balls.push_back(Ball);
    }
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Generator::UpdatePhysics -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Computes the ball positions and velocities for the next frame.
// By definition, the time of a frame is 1 unit of time.
//
void Generator::UpdatePhysics()
{
    double Time{0};                 // Current time.
    for (;;) {
        // Compute the time until a ball hits the wall or collides with another ball.
        auto [BallIdx, TimeToHitWall] = ComputeTimeToHitWall();
        auto [BallIdx1, BallIdx2, TimeToCollision] = ComputeTimeToCollision();
        double TimeToHit{min(TimeToHitWall, TimeToCollision)};

        // If the hit event occurs after the next frame time, we can break the loop and
        // draw the frame.
        if (Time + TimeToHit >= 1.0)
            break;

        // Move all balls to the position of the next hit.
        MoveBalls(TimeToHit);
        Time += TimeToHit;

        // Apply the velocity change coming from the wall hit or the ball collision
        if (TimeToHitWall < TimeToCollision)
            Balls[BallIdx].BounceAgainstWall();
        else
            Balls[BallIdx1].ComputeNewVelocities(Balls[BallIdx2]);
    }

    MoveBalls(1.0 - Time);
}

// -.-.-.-.-.-.-.-.-.-.-.- FrameGenerator::ComputeTimeToCollision -.-.-.-.-.-.-.-.-.-.-.-.
//
// Calculates the time until the first upcoming collision between any two balls.
// 
// Iterates through all possible pairs of balls in the frame, computing the time
// until each pair would collide based on their current velocities and positions. 
// It then identifies the minimum of these times.
//  
// Returns a triple containing the indices of the two balls that will collide first 
// and the duration until that collision.
//
tuple<int, int, double> Generator::ComputeTimeToCollision()
{
    double MinTimeToCollide{DBL_MAX};
    int BallIdx1{}, BallIdx2{};
    for (int i{0}; i<Balls.size()-1; i++)
        for (int j{i+1}; j<Balls.size(); j++)
        {
            double TimeToCollide{Balls[i].ComputeTimeToCollide(Balls[j])};
            if (TimeToCollide < MinTimeToCollide)
                MinTimeToCollide=TimeToCollide, BallIdx1=i, BallIdx2=j;
        }
    return {BallIdx1, BallIdx2, MinTimeToCollide};
}

// .-.-.-.-.-.-.-.-.-.-.-.- FrameGenerator::ComputeTimeToHitWall -.-.-.-.-.-.-.-.-.-.-.-.-
//
// Calculates the time until the first upcoming collision between any ball and one of
// the frame walls.
// 
// Iterates through all balls in the frame, computing the time until each would collide 
// with a wall based on its current velocity and position. It then identifies the minimum 
// of these times.
// 
// Returns a pair where 'BallIdx' is the index of the next ball to collide and 'Time'
// is the duration until that collision.
//
pair<int, double> Generator::ComputeTimeToHitWall()
{
    double MinTimeToHit{DBL_MAX};
    int BallIdx{0}, MinBallIdx{-1};
    for (const MovingBall& Ball : Balls)
    {
        double TimeToHit{Ball.ComputeTimeToHitWall()};
        if (TimeToHit < MinTimeToHit)
            MinBallIdx=BallIdx, MinTimeToHit=TimeToHit;
        BallIdx++;
    }
    return {MinBallIdx, MinTimeToHit};
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- FilledCircle::Draw -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Draws a filled circle on an image at the specified center point.
// This function uses the internal Pixels cache to draw fast.
// 
void FilledCircle::Draw(Image& Img, Point Center)
{
    Point UpperLeft{Center.x-r, Center.y-r};
    int Offset{3 * (UpperLeft.y*Img.b.w + UpperLeft.x)};

    uint8_t* CircY{Pixels.data()};
    uint8_t* YuvOut{Img.Pixels.data() + Offset};

    for (int y{0}; y<b.h; y++)
    {
        for (int x{0}; x<b.w; x++, CircY++, YuvOut+=3)
            if (*CircY == 255)
            {
                YuvOut[0] = c.Y;
                YuvOut[1] = c.U;
                YuvOut[2] = c.V;
            }
            else if (*CircY != 0)
            {
                double Alpha = CircY[0] / 255.0;
                YuvOut[0] = (uint8_t)(Alpha*c.Y + (1-Alpha)*YuvOut[0]);
                YuvOut[1] = (uint8_t)(Alpha*c.U + (1-Alpha)*YuvOut[1]);
                YuvOut[2] = (uint8_t)(Alpha*c.V + (1-Alpha)*YuvOut[2]);
            }
        YuvOut += 3 * (Img.b.w - b.w);
    }
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- FilledCircle::Prepare -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Prepares a filled circle for efficient drawing: caches the pixels of a supersampled 
// circle. Although the 10x supersampling used is not the most efficient, it simplifies 
// the code. Given this preparation step is not part of the real-time operation, the 
// efficiency trade-off is acceptable.
//
void FilledCircle::Prepare(int Radius, Color Col)
{
    r = Radius;
    c = Col;
    b = {r*2 + 1, r*2 + 1};
    Pixels.resize(b.w * b.h);

    int Super(10);
    uint8_t* Yout{Pixels.data()};
    for (int y{0}; y<2*r+1; y++)
        for (int x{0}; x<2*r+1; x++)
        {
            int Inside{0};
            for (int sy{-Super+1}; sy<=Super-1; sy+=2)
                for (int sx{-Super+1}; sx<=Super-1; sx+=2)
                {
                    int dx{r*Super*2 - (x*Super*2 + sx)};
                    int dy{r*Super*2 - (y*Super*2 + sy)};
                    if (dx*dx + dy*dy < r*Super*r*Super*4)
                        Inside++;
                }
            *Yout++ = (uint8_t)(255.0 * Inside / (Super*Super));
        }
}

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Image::Clear -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Fills the plane with a single color.
//
void Image::Clear(Color c)
{
    uint8_t* Yuv{Pixels.data()};
    for (int y{0}; y<b.h; y++)
        for (int x{0}; x<b.w; x++)
        {
            *Yuv++ = c.Y;
            *Yuv++ = c.U;
            *Yuv++ = c.V;
        }
}

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Image::ConvertToUyvy10 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Converts 8-bit YUV format to 10-bit packed UYVY.
//
void Image::ConvertToUyvy10(uint8_t* Uyvy10)
{
    uint8_t* Yuv8{Pixels.data()};
    for (int y{0}; y<b.h; y++)
    {
        for (int Sym{b.w}; Sym>0; Sym-=2)
        {
            // Convert 8-bit YUV YUV to 10-bit UYVY UYVY.
            uint32_t Val{0};
            Val = (Yuv8[1] + Yuv8[4]) << 1;
            Val |= Yuv8[0] << 12;
            Val |= (Yuv8[2] + Yuv8[5]) << 21;
            *(uint32_t*)Uyvy10 = Val;
            Uyvy10 += 4;
            *Uyvy10++ = Yuv8[3];
            Yuv8 += 6;
        }
    }
}
