// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* Generator.h *#*#*#*#*#*#*#*#*#*#* (C) 2023 DekTec
//
// Code example AvFifo_VideoTx - Moving balls frame generator - Header
//
// Note:
// - Although this file is part of the AvFifo_VideoTx code example, it is primarily 
//   concerned with the pattern generation aspect, and does not directly interact with  
//   the AvFifo API.
// 
// This header file defines the Generator class which is used for generating a simulation
// of moving and colliding balls within a frame. It is designed to visualize a simple 
// physics system in a 2D space, to create a dynamic and engaging test pattern.
//
// The Generator class provides methods to initialize the simulation, advance the 
// simulation, compute next frames, and generate output video frames. It also contains 
// methods to calculate time to collision between balls and between a ball and a wall.
//
// This simulation assumes that all collisions are perfectly elastic and that the balls
// have uniform properties such as mass and radius.
//
// The output video frames are formatted in Uyvy10 format, which matches the format
// expected by the SMPTE 2110 video standard.
//
// The initial positions, velocities, and colors of the balls are randomly chosen, based
// on the parameters in 'Configuration.h'. The size of the simulation frame is also
// set in 'Configuration.h'.
//

#pragma once

// -.-.-.-.-.-.-.-.-.-.-.- Include necessary headers and libraries -.-.-.-.-.-.-.-.-.-.-.-

#include <cfloat>                   // Standard library: Floating point numeric limits.
#include <cmath>                    // Standard library: Math operations, for sqrt().
#include <iostream>                 // Standard library: Console input/output operations.
#include <tuple>                    // Standard library: Tuple for grouping elements.
#include <vector>                   // Standard library: Dynamic-size array handling.

using namespace std;                // Avoid std:: prefix for standard library calls.

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.- Elementary data structures -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

struct Color {uint8_t Y, U, V;};    // YUV color representation.
struct Coordinate {double x, y;};   // 2D floating-point coordinate.
struct Point {int x, y;};           // 2D integer point.
struct Size {int w, h;};            // 2D integer size.
struct Velocity {double x, y;};     // 2D velocity vector.

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.- Intermediate helper structs -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
 
// Represents a pixel-centered circle with radius r and bounding box size 2*r+1. 
// Use Prepare() for caching pixels, Draw() for rendering on a plane.
struct FilledCircle
{
    Size b{};                       // Bounding box, size = 2*r+1.
    Color c;                        // Circle inner color.
    vector<uint8_t> Pixels{};       // Cached circle pixels.
    int r{};                        // Circle radius.

    void Draw(struct Image&, Point);// Draws circle at a given center point.
    void Prepare(int r, Color c);   // Prepares circle pixels for efficient drawing.
};

// Represents a 2D image with 24-bit (3x 8-bit) YUV pixels.
struct Image
{
    Size b{};                       // Size of the image.
    vector<uint8_t> Pixels;         // YUV pixels.

    void Clear(Color c);            // Clears image with a specific color.
    void ConvertToUyvy10(uint8_t* Uyvy10);
    void Init(Size s) {b = s; Pixels.resize(3 * b.w * b.h);}
};

// .-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- struct MovingBall -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Encapsulates a moving ball, and its operations.
//
struct MovingBall
{
    Color c;                        // Color of the ball.
    double m{};                     // Mass, used in post-collision velocity calculations.
    Size MovingBounds{};            // Bounding box within which the balls can move.
    Coordinate p{};                 // Position of the center of the ball.
    int r{};                        // Radius of the wall.
    Velocity v{};                   // Velocity of the ball.

    // Does this ball overlap with another one?
    bool DoesOverlap(const MovingBall& Other) const
    {
        double dx{p.x-Other.p.x}, dy{p.y-Other.p.y}, rSum{(double)r + Other.r};
        return dx*dx + dy*dy < rSum*rSum;
    }

    // Checks if ball fits within a box.
    bool FitsInBox(Size Box) const
    {
        return p.x-r>=0 && p.x+r<=Box.w && p.y-r>=0 && p.y+r<=Box.h;
    }

    // Calculates new velocities after collision.
    // A bit of classical mechanics: preservation of momentum and energy.
    void ComputeNewVelocities(MovingBall& Other)
    {
        Coordinate pRel{Other.p.x-p.x, Other.p.y-p.y};
        Velocity vRel1{v.x-Other.v.x, v.y-Other.v.y};
        Velocity vRel2{Other.v.x-v.x, Other.v.y-v.y};

        double MagSq{pRel.x*pRel.x + pRel.y*pRel.y};
        double f1{2*Other.m/(m+Other.m) * (vRel1.x*pRel.x+vRel1.y*pRel.y) / MagSq};
        double f2{2*m/(m+Other.m) * (-vRel2.x*pRel.x-vRel2.y*pRel.y) / MagSq};

        v = {v.x-f1*pRel.x, v.y-f1*pRel.y};
        Other.v = {Other.v.x+f2*pRel.x, Other.v.y+f2*pRel.y};
    }

    // Draws the moving ball by using the already rendered and cached pixels.
    void Draw(Image& Yuv) {Circle.Draw(Yuv, {(int)p.x, (int)p.y});};

    // Adjusts the ball's position based on its velocity and the given amount of time.
    void Move(double DeltaT) {p.x += v.x*DeltaT; p.y += v.y*DeltaT;}

    // Prepares this ball for rendering by caching the pixels.
    void Prepare(Color Col) {c=Col; Circle.Prepare(r, c);}

    // Bounces this ball against the frame wall.
    void BounceAgainstWall()
    {
        if ((p.x-r<=0.0 && v.x<0.0) || (p.x+r>=MovingBounds.w && v.x>0.0))
            v.x = -v.x;
        if ((p.y-r<=0.0 && v.y<0.0) || (p.y+r>=MovingBounds.h && v.y>0.0))
            v.y = -v.y;
    }

    // Computes the time it will take for this ball to collide with another ball.
    double ComputeTimeToCollide(const MovingBall& Other) const
    {
        // Relative position and velocity.
        Velocity vRel{v.x-Other.v.x, v.y-Other.v.y};
        Coordinate pRel{p.x-Other.p.x, p.y-Other.p.y};

        // Computing collision time means solving a quadratic equation, which has zero
        // or two solutions. We are interested in the earliest collision.
        double a{vRel.x*vRel.x + vRel.y*vRel.y};
        double b{2 * (pRel.x*vRel.x + pRel.y*vRel.y)};
        double c{pRel.x*pRel.x + pRel.y*pRel.y - (r+Other.r)*(r+Other.r)};
        double Discriminant{b*b - 4*a*c};
        if (Discriminant < 0) 
            return DBL_MAX;         // The balls do not collide.
        double t1{(-b - sqrt(Discriminant)) / (2*a)};
        double t2{(-b + sqrt(Discriminant)) / (2*a)};
        double CollisionTime{min(t1, t2)};
    
        // If CollisionTime is negative, the collision is in the past. 
        // This means the balls are moving away from each other.
        return CollisionTime>=0.0 ? CollisionTime : DBL_MAX;
    }

    // Computes the time it will take for this ball to hit a frame boundary.
    double ComputeTimeToHitWall() const
    {
        double DistToWallX{DBL_MAX}, DistToWallY{DBL_MAX};
        if (v.x < 0.0)
            DistToWallX = p.x - r;
        else if (v.x > 0.0)
            DistToWallX = MovingBounds.w - (p.x + r);
        if (v.y < 0.0)
            DistToWallY = p.y - r;
        else if (v.y > 0.0)
            DistToWallY = MovingBounds.h - (p.y + r);

        double TimeToHitX{v.x!=0.0 ? DistToWallX/abs(v.x) : DBL_MAX};
        double TimeToHitY{v.y!=0.0 ? DistToWallY/abs(v.y) : DBL_MAX};

        return min(TimeToHitX, TimeToHitY);
    }

private:
    FilledCircle Circle;            // Caches the circle's pixels for fast drawing.
};

// -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- class Generator -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Handles the generation of moving ball simulation and rendering of video frames.
//
class Generator
{
public:
    // Computes the next simulation frame, assuming 20ms time increments.
    void UpdatePhysics();

    // Generates video frame in Uyvy10 format.
    void Generate(uint8_t* Uyvy10);

    // Initializes the simulation with random ball positions, velocities, and colors.
    void Initialize(Size FrameSize);

    // Advances the simulation by a given time step.
    // This method assumes no collisions will occur during the given time step.
    void MoveBalls(double DeltaT) {for (MovingBall& Ball : Balls) Ball.Move(DeltaT);}

private:
    vector<MovingBall> Balls;       // Collection of moving balls within the frame.
    Size FrameSize;                 // Dimensions of the simulation frame.
    Image Img;                      // Image object for rendering balls.

    // Returns the indices of two colliding balls and the time until their collision.
    tuple<int, int, double> ComputeTimeToCollision();

    // Returns the index of a ball and the time it takes for the ball to hit a wall.
    pair<int, double> ComputeTimeToHitWall();
};
